import numpy as np
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from itertools import chain
import time
import pathlib

from repeater_chain_analyzer import RepeaterChainCalculator
import plotting_tools as plot_tools


def calculate_data():
    pgens = np.linspace(0,1, 41)[1:]
    pswap = 1
    n_max = 4
    trunc = int(500)

    for pgen in pgens:
        print("running pgen={}".format(pgen))
        datafolder = 'results/7a/data/pgen{}/'.format(round(pgen,3))

        calculator = RepeaterChainCalculator(n=n_max,
                                             trunc=trunc,
                                             pgen=pgen,
                                             pswap=pswap,
                                             outputfolder=datafolder)
        calculator.calculate()

    #plot_tools.plot_distributions(outputfolder='results/7a/plots/',
    #                              rca=calculator,
    #                              to_plot='pmf',
    #                              level_selection=[3,4])

def calculate_data_extended():
    pgens = np.linspace(0,1, 41)[1:]
    pswap = 1
    n_max = 11
    trunc = int(1000)

    for pgen in pgens:
        print("running pgen={}".format(pgen))
        datafolder = 'results/7a_extended/data/pgen{}/'.format(round(pgen,3))

        calculator = RepeaterChainCalculator(n=n_max,
                                             trunc=trunc,
                                             pgen=pgen,
                                             pswap=pswap,
                                             outputfolder=datafolder)
        calculator.calculate()

def plot_7a():
    # hardcoded parameters
    n_max = 4
    pswap = 1
    pgens = np.linspace(0,1, 41)[1:]

    # gather data
    means_lower = np.zeros(shape=(n_max+1,len(pgens)))
    means_upper = np.zeros(shape=(n_max+1,len(pgens)))
    means3over2 = np.ones(shape=(n_max+1,len(pgens)))

    for i, pgen in enumerate(pgens):
        datafolder = 'results/7a/data/pgen{}/'.format(round(pgen,3))
        calculator = RepeaterChainCalculator(n=0, pgen=0, pswap=0, trunc=0)
        calculator.load_from_folder(datafolder)

        for n in range(1, n_max+1):
            lower, upper = calculator.mean_bounds(n)
            means_lower[n,i] = lower
            means_upper[n,i] = upper
            means3over2[n,i] = (3 / (2*pswap))**n * (1 / pgen)

    ratio_lower = means_lower / means3over2
    ratio_upper = means_upper / means3over2

    # plot stuff
    fig, ax = plt.subplots(1,1)

    scaling = 4.25 # this mostely changes the 'zoom level' / 'scale'
    width  = scaling
    height = scaling * 0.65
    fig.set_figwidth(width)
    fig.set_figheight(height)

    default_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    col_counter = 0
    legend_lables = []
    legend_handles = []

    for n in range(1, n_max+1):
        color = default_colors[col_counter]
        col_counter += 1
        ax.plot(pgens, ratio_upper[n], color=color)
        ax.plot(pgens, ratio_lower[n], color=color)
        legend_lables.append('$N = {}$'.format(2**n))
        legend_handles.append(mlines.Line2D([],[],color=color))


    ax.legend(labels=legend_lables, handles=legend_handles)
    ax.set(xlabel='$p_{gen}$')
    ax.set(ylabel='computed mean / 3-over-2')
    plt.savefig('results/7a/plots/ratio_plot.eps',bbox_inches='tight')


def plot_7a_extended():
    # hardcoded parameters
    n_max = 11
    pswap = 1
    pgens = np.linspace(0,1, 41)[1:]
    upper_plot = [1,2,3,4]
    lower_plot = [9,10,11]

    # gather data
    means_lower = np.zeros(shape=(n_max+1,len(pgens)))
    means_upper = np.zeros(shape=(n_max+1,len(pgens)))
    means3over2 = np.ones(shape=(n_max+1,len(pgens)))

    for i, pgen in enumerate(pgens):
        datafolder = 'results/7a_extended/data/pgen{}/'.format(round(pgen,3))
        calculator = RepeaterChainCalculator(n=0, pgen=0, pswap=0, trunc=0)
        calculator.load_from_folder(datafolder)

        for n in range(1, n_max+1):
            lower, upper = calculator.mean_bounds(n)
            means_lower[n,i] = lower
            means_upper[n,i] = upper
            means3over2[n,i] = (3 / (2*pswap))**n * (1 / pgen)

    ratio_lower = means_lower / means3over2
    ratio_upper = means_upper / means3over2

    # plot stuff
    subplots = True
    if(not subplots):
        fig, ax = plt.subplots(1,1)
        scaling = 4.25 # this mostely changes the 'zoom level' / 'scale'
        width  = scaling
        height = scaling * 1
        fig.set_figwidth(width)
        fig.set_figheight(height)

        default_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        col_counter = 0
        legend_lables = []
        legend_handles = []

        for n in chain(upper_plot, lower_plot):
            color = default_colors[col_counter]
            col_counter += 1
            ax.plot(pgens, ratio_upper[n], color=color)
            ax.plot(pgens, ratio_lower[n], color=color)
            legend_lables.append('$N = {}$'.format(2**n))
            legend_handles.append(mlines.Line2D([],[],color=color))
        ax.legend(labels=legend_lables, handles=legend_handles)

        ax.set(xlabel='$p_{gen}$')
        ax.set(ylabel='computed mean / 3-over-2')
        plt.savefig('results/7a_extended/plots/ratio_plot.eps',bbox_inches='tight')
    elif(subplots):
        fig, axs = plt.subplots(2,1,
                                sharex='col',
                                gridspec_kw={'hspace': 0, 'height_ratios': [1.25, 1]})

        scaling = 4.25 # this mostely changes the 'zoom level' / 'scale'
        width  = scaling
        height = scaling * 1
        fig.set_figwidth(width)
        fig.set_figheight(height)

        default_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        col_counter = 0
        legend_lables = []
        legend_handles = []

        for n in upper_plot:
            ax = axs[0]
            color = default_colors[col_counter]
            col_counter += 1
            ax.plot(pgens, ratio_upper[n], color=color)
            ax.plot(pgens, ratio_lower[n], color=color)
            legend_lables.append('$N = {}$'.format(2**n))
            legend_handles.append(mlines.Line2D([],[],color=color))
            ax.legend(labels=legend_lables, handles=legend_handles)

        default_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        col_counter = 0
        legend_lables = []
        legend_handles = []

        for n in lower_plot:
            ax = axs[1]
            color = default_colors[col_counter]
            col_counter += 1
            ax.plot(pgens, ratio_upper[n], color=color)
            ax.plot(pgens, ratio_lower[n], color=color)
            legend_lables.append('$N = {}$'.format(2**n))
            legend_handles.append(mlines.Line2D([],[],color=color))
            ax.legend(labels=legend_lables, handles=legend_handles)
        axs[1].set(yticks=[0, 0.05, 0.10, 0.15])
        #axs[1].set(yticks=[0,0.1])

        axs[1].set(xlabel='$p_{gen}$')
        #axs[0].set(ylabel='computed mean / 3-over-2')
        fig.text(0, 0.5, 'computed mean / 3-over-2', ha='center', va='center', rotation='vertical')
        plt.savefig('results/7a_extended/plots/ratio_plot.eps',bbox_inches='tight')

if __name__ == '__main__':
    pathlib.Path('results/7a_extended/plots').mkdir(parents=True, exist_ok=True)
    start = time.time()
    calculate_data_extended()
    runtime = time.time() - start
    print(runtime)
    plot_7a_extended()
