import numpy as np
import matplotlib.pyplot as plt
from itertools import chain
from scipy.stats import gumbel_r, genextreme
from scipy.stats import ks_2samp
from scipy.optimize import curve_fit
import pathlib
import json

from repeater_chain_analyzer import RepeaterChainCalculator
import plotting_tools as plot_tools
import probability_tools as prob_tools


def generate_data(folder):
    # parameters
    n_max   = 8
    trunc   = int(1.9e5)
    pgen    = 0.1
    pswap   = 0.5

    # Calculate data
    calculator = RepeaterChainCalculator(n=n_max,
                                         trunc=trunc,
                                         pgen=pgen,
                                         pswap=pswap,
                                         outputfolder=folder)
    calculator.calculate()

def fit_and_plot(datafolder, func, bounds, name):
    # parameters
    n_train = range(1,5)
    n_test  = range(5,9)
    outputfolder = "results/{}/".format(name)
    outputpath = outputfolder + "fit_n{}.eps"
    pathlib.Path(outputfolder).mkdir(parents=True, exist_ok=True)

    # Load
    calculator = RepeaterChainCalculator(n=0, pgen=0, pswap=0, trunc=0)
    calculator.load_from_folder(datafolder)
    params = calculator.params
    with open(outputfolder + "repeater_params.json", "w") as f:
        json.dump(params, f)

    cols = plt.rcParams['axes.prop_cycle'].by_key()['color']

    popts = []
    for n in chain(n_train, n_test):
        print("fitting for n={}".format(n))
        # for this `n`, find optimal pos and scale values (up to CDF = 0.99)
        pmf = np.copy(calculator.pmf[n])
        cdf = prob_tools.pmf_to_cdf(pmf)
        pmf = pmf[np.where(cdf < 0.99)] # truncate pmf to 0.99 mass

        # fit parameters of func to best aproximate numerical pmf
        new_trunc = len(pmf)
        x = range(new_trunc)
        popt, pcov = curve_fit(func, xdata=x, ydata=pmf, bounds=bounds)
        popts.append(popt)

        # plot
        y_axis_factor = 10**4 # NOTE: hardcoded in label
        fig, ax = plt.subplots(1,1)
        scaling = 4.25 # this mostely changes the 'zoom level' / 'scale'
        width  = scaling
        height = scaling * 0.65
        fig.set_figwidth(width)
        fig.set_figheight(height)

        ax.step(x, pmf*y_axis_factor, color=cols[0])
        ax.plot(x, func(x, *popt)*y_axis_factor, color=cols[1], linestyle='--')
        # means
        left, right = calculator.mean_bounds(n)
        right = min(int(np.round(right)), new_trunc)
        left = int(np.round(left))
        if(left <= right):
            plt.fill_between(x=range(left,right+1),
                             y1=pmf[left:right+1] * y_axis_factor,
                             color=cols[0], alpha=0.5)
        # technical bebt:
        if(name == 'GEV' and n <=6):
            gev_mean = genextreme.stats(c=popt[0],
                                        loc=popt[1],
                                        scale=popt[2],
                                        moments='m')
            ymax = func(gev_mean, *popt)
            ax.vlines(x=gev_mean, ymin=0, ymax=ymax*y_axis_factor, color=cols[1])
        # title and stuff
        ax.set_ylim(bottom=0)
        ax.set(ylabel='$\\Pr(T_n = t)$ $(10^{-4})$')
        ax.set(xlabel='$t$')
        #ax.set(yticks=[0, 0.3, 0.6, 0.9, 1.2, 1.5])
        ax.set(yticks=[0, 0.4, 0.8, 1.2, 1.6])
        #ax.set(yticks=[0, 0.5, 1, 1.5])
        legend = []
        legend.append('deterministic algorithm'.format(2**n))
        legend.append('fitted distribution'.format(name, round(popt[0], 4)))
        ax.legend(legend)
        #fig.set_figwidth(width)
        #fig.set_figheight(height)
        fig.savefig(outputpath.format(n), dpi=300, bbox_inches='tight')
        fig.clf()

    np.savetxt(outputfolder + "fit_params.csv", popts, delimiter=',')


def fit_and_plot_gumbel(datafolder):
    """ Fits a Gumbel distribution through the numerical pmf """

    # we need a wrapper because of the way `curve_fit` works
    def func(x, loc, scale):
        return gumbel_r.pdf(x, loc=loc, scale=scale)

    # bounds on parameter search space [loc, scale]
    lower_bounds = [-np.inf, -np.inf]
    upper_bounds = [ np.inf,  np.inf]
    bounds = (lower_bounds, upper_bounds)

    return fit_and_plot(datafolder, func, bounds, 'gumbel')


def fit_and_plot_GEV(datafolder):
    """ Fits a GEV distribution through the numerical pmf """

    # for c = 0 this is a Gumbel distribution
    def func(x, c, loc, scale):
        return genextreme.pdf(x, c=c, loc=loc, scale=scale)

    # bounds on parameter search space [c, loc, scale]
    lower_bounds = [-1, -np.inf, -np.inf]
    upper_bounds = [0, np.inf, np.inf]
    bounds = (lower_bounds, upper_bounds)

    return fit_and_plot(datafolder, func, bounds, 'GEV')


if __name__ == '__main__':

    datafolder  = 'results/data_for_GEV_fit/'
    generate_data(datafolder)

    #fit_and_plot_gumbel(datafolder)
    fit_and_plot_GEV(datafolder)


















#
